#include <vector>
#include <set>
#include <algorithm>
#include <iostream>
#include <stack>
#include "formula.hpp"
#include "aux.hpp"

/**
 * \file formula.cpp
 * \brief Implementation of basic functions for variable, clause, and dependency management
 * \author Ralf Wimmer
 * \date 12/2015-03/2016
 */


/**
 * \brief Creates a new universal variable.
 *
 * \param index the ID of the created variable
 * \pre The variable may not already occur in one of the clauses.
 * \sa Formula::addEVar
 */
void Formula::addUVar(unsigned int index)
{
    val_assert(index > maxVarIndex() || varDeleted(index));

    if (index > maxVarIndex()) {
        var_status.resize(index + 1, VariableStatus::DELETED);
        prefix.resize(index + 1);
        occ_list.resize(2 * index + 2);
        implications.resize(2 * index + 2);
        assignment.resize(index, TruthValue::UNKNOWN);
    }

    var_status[index] = VariableStatus::UNIVERSAL;
    ++num_u_vars;
}


/**
 * \brief Returns the accumulated number of dependencies of all existential variables.
 */
unsigned int Formula::numDependencies() const noexcept
{
    unsigned int num = 0;
    for (unsigned int var = 1; var < prefix.size(); ++var) {
        if (isExistential(var)) num += prefix[var].size();
    }
    return num;
}


/**
 * \brief Removes all dependencies of a given variable.
 */
void Formula::clearDependencies(unsigned int var)
{
    for (unsigned int dep: prefix[var]) prefix[dep].erase(var);
    prefix[var].clear();
}


/**
 * \brief Checks if `var1`'s dependencies are a subset of `var2`'s dependencies.
 *
 * If `var1` is universal, this function returns whether `var2` depends on `var1`.
 * If both variables are existential, this function checks whether the dependency
 * set of `var1` is a subset of `var2`'s dependencies.
 *
 * \note `var2` has to be existential.
 */
bool Formula::dependenciesSubset(unsigned int var1, unsigned int var2) const
{
    val_assert(var1 <= maxVarIndex() && var2 <= maxVarIndex());
    val_assert(!varDeleted(var1) && !varDeleted(var2));

    if (var1 == var2) return true;
    if (isUniversal(var1) && isUniversal(var2)) return false;
    if (isUniversal(var1)) return depends(var2, var1);
    if (isUniversal(var2)) return prefix[var1].empty();
    if (prefix[var2].size() == num_u_vars) return true;
    if (prefix[var1].size() > prefix[var2].size()) return false;

    return std::includes(prefix[var2].cbegin(), prefix[var2].cend(),
                         prefix[var1].cbegin(), prefix[var1].cend());
}


/**
 * \brief Adds a clause to the formula.
 *
 * \note Calling this function is only allowed as long as none of
 *       the functions has been called which modify the formula
 *       (like unitPropagation(), findPure(), ...)
 * \param _clause vector with the literals of the clause
 * \param needs_sorting if true, the clause is sorted and duplicate literals are removed.
 * \pre If needs_sorting is false, the caller has to guarantee that the literals are
 *      ordered in increasing order and that no literal appears more than once.
 * \return the clause ID if the clause was actually added; -2 if the clause was a tautology; -1 if the clause was unit.
 * \throw UNSATException if an empty clause is added.
 */
int Formula::addClause(const std::vector<unsigned int>& _clause, bool needs_sorting)
{
    if (_clause.empty()) {
        throw UNSATException("Trying to add an empty clause");
    }

    unsigned int c_nr = 0;

    // Try to recycle clause numbers where possible.
    if (deleted_clause_numbers.empty()) {
        clauses.push_back(_clause);
        c_nr = clauses.size() - 1;
    } else {
        c_nr = deleted_clause_numbers.back();
        deleted_clause_numbers.pop_back();
        clauses[c_nr] = _clause;
    }
    auto& clause = clauses[c_nr];

    // We have to remove duplicate literals and test for tautological clauses
    if (needs_sorting) {
        std::sort(clause.begin(), clause.end());
        clause.erase(std::unique(clause.begin(), clause.end()), clause.end());
    }

    for (unsigned int pos = 0; pos < clause.size() - 1; ++pos) {
        val_assert(clause[pos] != clause[pos+1]);
        if (lit2var(clause[pos]) == lit2var(clause[pos+1]) && clause[pos] != clause[pos+1]) {
            // Clause is a tautology. We can ignore it.
#ifdef DEBUG_PRINT
            std::cout << __FUNCTION__ << "(copy) Ignoring tautology " << PrintClause(clause) << std::endl;
#endif
            clauses[c_nr].clear();
            clause_status[c_nr] = ClauseStatus::DELETED;
            deleted_clause_numbers.push_back(c_nr);
            return -2;
        }
    }

    // fill occurrence lists
    for (unsigned int lit: clause) {
        occ_list[lit].push_back(c_nr);
    }

    // for binary clauses, add implications
    if (clause.size() == 2) {
        // binary clauses are additionally stored as implications.
        addImplication(negate(clause[0]), clause[1]);
        addImplication(negate(clause[1]), clause[0]);
    }

    if (clause_status.size() <= c_nr) clause_status.resize(c_nr + 1, ClauseStatus::DELETED);
    clause_status[c_nr] = ClauseStatus::MANDATORY;

    return c_nr;
}


/**
 * \brief Adds a clause to the formula.
 *
 * \note Calling this function is only allowed as long as none of
 *       the functions has been called which modify the formula
 *       (like unitPropagation(), findPure(), ...)
 * \param _clause vector with the literals of the clause
 * \param needs_sorting if true, the clause is sorted and duplicate literals are removed.
 * \pre If needs_sorting is false, the caller has to guarantee that the literals are
 *      ordered in increasing order and that no literal appears more than once.
 * \return the clause ID if the clause was actually added; -1 if the clause was a tautology.
 * \throw UNSATException if an empty clause is added.
 */
int Formula::addClause(std::vector<unsigned int>&& _clause, bool needs_sorting)
{
    if (_clause.empty()) {
        throw UNSATException("Trying to add an empty clause");
    }

    unsigned int c_nr = 0;
    // Try to recycle clause numbers where possible.
    if (deleted_clause_numbers.empty()) {
        clauses.push_back(std::move(_clause));
        c_nr = clauses.size() - 1;
    } else {
        c_nr = deleted_clause_numbers.back();
        deleted_clause_numbers.pop_back();
        clauses[c_nr] = std::move(_clause);
    }
    auto& clause = clauses[c_nr];

    // We have to remove duplicate literals and test for tautological clauses
    if (needs_sorting) {
        std::sort(clause.begin(), clause.end());
        clause.erase(std::unique(clause.begin(), clause.end()), clause.end());
    }

    for (unsigned int pos = 0; pos < clause.size() - 1; ++pos) {
        val_assert(clause[pos] != clause[pos+1]);
        if (lit2var(clause[pos]) == lit2var(clause[pos+1]) && clause[pos] != clause[pos+1]) {
            // Clause is a tautology. We can ignore it.
#ifdef DEBUG_PRINT
            std::cout << __FUNCTION__ << "(move) Ignoring tautology " << PrintClause(clause) << std::endl;
#endif
            clauses[c_nr].clear();
            clause_status[c_nr] = ClauseStatus::DELETED;
            deleted_clause_numbers.push_back(c_nr);
            return -2;
        }
    }

    // fill occurrence lists
    for (unsigned int lit: clause) {
        occ_list[lit].push_back(c_nr);
    }

    // for binary clauses, add implications
    if (clause.size() == 2) {
        // binary clauses are additionally stored as implications.
        addImplication(negate(clause[0]), clause[1]);
        addImplication(negate(clause[1]), clause[0]);
    }

    if (clause_status.size() <= c_nr) clause_status.resize(c_nr + 1, ClauseStatus::DELETED);
    clause_status[c_nr] = ClauseStatus::MANDATORY;

    return c_nr;
}


/**
 * \brief Removes a clause from the formula.
 *
 * \param c_nr the ID of the clause
 * \note This function modifies Formula::occ_list and Formula::implications.
 *       Therefore Formula::removeClause may not be called while iterating over
 *       one of these lists.
 * \note The removed clause may not be accessed anymore.
 */
void Formula::removeClause(const unsigned int c_nr)
{
    val_assert(c_nr < clauses.size());
    if (clauseDeleted(c_nr)) {
        val_assert(clauses[c_nr].empty());
        return;
    }

    for (unsigned int lit: clauses[c_nr]) {
        // remove c_nr from occ_list[lit]:
        // - find the position of 'lit'
        // - copy the last element of 'occ_list[lit]' to 'lit's position
        // - drop the last element.
        auto position = std::find(occ_list[lit].begin(), occ_list[lit].end(), c_nr);
        val_assert(position != occ_list[lit].end());
        *position = occ_list[lit].back();
        occ_list[lit].pop_back();
    }

    if (clauses[c_nr].size() == 2) {
        // remove both implications
        removeImplication(negate(clauses[c_nr][0]), clauses[c_nr][1]);
        removeImplication(negate(clauses[c_nr][1]), clauses[c_nr][0]);
    }

    // clear the clause and free its number
    clauses[c_nr].clear();
    clause_status[c_nr] = ClauseStatus::DELETED;
    deleted_clause_numbers.push_back(c_nr);
}


/**
 * \brief Resets the formula and clears all the data.
 *
 * Afterwards, the list of variables and clauses as well as
 * all auxiliary data structures (implications, occurrence lists)
 * are empty. The statistics variables are reset to 0.
 */
void Formula::reset()
{
    var_status.clear();
    clause_status.clear();
    prefix.clear();
    clauses.clear();
    occ_list.clear();
    implications.clear();
    deleted_clause_numbers.clear();
    unit_stack.clear();
    assignment.clear();
    num_u_vars = 0;
    num_e_vars = 0;
}


/**
 * \brief Checks if start_lit implies target_var or NOT(target_var) taking transitivity into account.
 *
 * \return a pair (a,b) such that a = true means start_lit implies target_var and
 *         b = true that start_lit implies NOT(target_var).
 */
std::pair<bool,bool> Formula::hasImplicationTransitive(unsigned int start_lit, unsigned int target_var) const
{
    val_assert(start_lit < implications.size());

    std::stack<unsigned int> front( {start_lit} );
    std::vector<bool> reached(implications.size(), false);
    reached[start_lit] = true;

    while (!front.empty()) {
        const unsigned int top = front.top();
        front.pop();
        for (unsigned int succ: implications[top]) {
            if (!reached[succ]) {
                reached[succ] = true;
                front.push(succ);
            }
        }
    }

    return std::make_pair(reached[var2lit(target_var, false)], reached[var2lit(target_var, true)]);
}

