#ifndef FORMULA_HPP
#define FORMULA_HPP

#include <vector>
#include <set>
#include <exception>
#include <ostream>
#include <string>
#include "aux.hpp"

/**
 * \file formula.hpp
 * \brief Main header file for DQBF formulas.
 * \author Ralf Wimmer
 * \date 2015-2016
 * \details Contains all functions for manipulating literals and DQBFs.
 */


//@{
/**\name Manipulation of literals
*/

/**
 * \brief Converts a literal to its underlying variable.
 */
inline constexpr unsigned int lit2var(const unsigned int lit) noexcept
{
    return (lit >> 1);
}


/**
 * \brief Converts a variable to a literal with the given polarity.
 */
inline constexpr unsigned int var2lit(const unsigned int var, const bool negated = false) noexcept
{
    return ((var << 1) | (negated ? 1 : 0));
}


/**
 * \brief Converts a literal to DIMACS format.
 *
 * Positive literals correspond to the variable index,
 * negative literals to the negative variable index.
 */
inline constexpr int lit2dimacs(const unsigned int lit) noexcept
{
    return ((lit & 1) == 1) ? (-1 * static_cast<int>(lit2var(lit))) : (static_cast<int>(lit2var(lit)));
}

/**
 * \brief Negates a given literal.
 */
inline constexpr unsigned int negate(const unsigned int lit) noexcept
{
    return (lit ^ 1);
}


/**
 * \brief Negates a given literal iff 'negate' is true.
 */
inline constexpr unsigned int negate_if(const unsigned int lit, const bool negate) noexcept
{
    return (negate ? (lit ^ 1) : lit);
}


/**
 * \brief Checks whether a given literal is positive.
 */
inline constexpr bool isPositive(const unsigned int lit) noexcept
{
    return ((lit & 1) == 0);
}


/**
 * \brief Checks whether a given literal is negative.
 */
inline constexpr bool isNegative(const unsigned int lit) noexcept
{
    return ((lit & 1) != 1);
}
//@}


/**
 * \brief Type of the dependency scheme that is to be applied.
 */
enum class DependencyScheme: int
{
    TRIVIAL = 0,        ///< Trivial dependency scheme as given in the prefix
    STANDARD,           ///< Standard dependency scheme
    STRICT_STANDARD,    ///< Strict standard dependency scheme
    REF_TRIANGLE,       ///< Reflexive triangle dependency scheme
    REF_QUADRANGLE,     ///< Reflexive quadrangle dependency scheme
    RP_STANDARD,        ///< Standard resolution path dependency scheme
    RP_STRICT_STANDARD, ///< Strict standard resolution path dependency scheme
    RP_REF_TRIANGLE,    ///< Reflexive triangle dependency scheme
    RP_REF_QUADRANGLE,  ///< Reflexive quadrangle dependency scheme
    GATE                ///< Gate detection
};


/**
 * \brief How to modify the dependencies
 */
enum class DependencyOperation: int
{
    ADD,       ///< Add dependencies using the selected dependency scheme
    REMOVE,    ///< Remove dependencies using the selected dependency scheme
    DO_NOTHING ///< Do not modify the dependencies.
};


/**
 * Represents the truth value of a variable or a formula.
 */
enum class TruthValue: int
{
    FALSE = 0,  ///< false or unsatisfiable
    TRUE = 1,   ///< true or satisfiable
    UNKNOWN = 2 ///< unassigned or unknown truth value
};


/**
 * \brief Exception that is thrown when the formula is recognized to be unsatisfiable.
 */
class UNSATException: public std::exception
{
public:
    UNSATException() = default;

    /**
     * \brief Creates a new UNSATException with a description of its reason
     * \param msg description of the exception's reaon
     */
    UNSATException(const char * msg): std::exception(), message(msg) { }

    /**
     * \brief Creates a new UNSATException with a description of its reason
     * \param msg description of the exception's reaon
     */
    UNSATException(const std::string& msg): std::exception(), message(msg) { }
    virtual ~UNSATException() = default;

    /**
     * \brief Returns a description of the reasons why the exception was thrown.
     */
    virtual const char* what() const noexcept override { return message.c_str(); }

private:
    std::string message; ///< Reason for the exception
};


/**
 * \brief Exception that is thrown when the formula is recognized to be satisfiable.
 */
class SATException: public std::exception
{
public:
    SATException() = default;

    /**
     * \brief Creates a new SATException with a description of its reason
     * \param msg description of the exception's reaon
     */
    SATException(const char * msg): std::exception(), message(msg) { }

    /**
     * \brief Creates a new SATException with a description of its reason
     * \param msg description of the exception's reaon
     */
    SATException(const std::string& msg): std::exception(), message(msg) { }
    virtual ~SATException() = default;

    /**
     * \brief Returns a description of the reasons why the exception was thrown.
     */
    virtual const char* what() const noexcept override { return message.c_str(); }

private:
    std::string message; ///< Reason for the exception
};


/**
 * \brief Type of a gate.
 */
enum class GateType
{
    AND_GATE, ///< AND-gate
    XOR_GATE, ///< XOR-gate
    UNKNOWN   ///< unknown (invalid) gate
};


/**
 * \brief Status of a variable (quantifier/deleted)
 */
enum class VariableStatus: short
{
    EXISTENTIAL, ///< Variable is existential
    UNIVERSAL,   ///< Variable is universal
    DELETED      ///< Variable has been deleted
};


/**
 * \brief Status of a clause (mandatory/optional/deleted)
 */
enum class ClauseStatus: short
{
    MANDATORY, ///< actual problem clause
    /**
     * \brief redundant clause which may be deleted if not used.
     *
     * Optional clauses must be implications from the mandatory clauses.
     * They can be deleted without changing the set of satisfying assignments
     * of the (mandatory) matrix.
     */
    OPTIONAL,
    DELETED    ///< deleted (invalid) clause
};


/**
 * \brief Representation of a gate extracted from a formula in CNF.
 */
class Gate
{
public:
    /**
     * \brief Constructs a gate without inputs and unknown type.
     */
    Gate():
        type(GateType::UNKNOWN),
        output_literal(0),
        input_literals(),
        encoding_clauses()
    { }

    GateType type;                              ///< Type of the gate (AND/XOR)
    unsigned int output_literal;                ///< output literal (might be negated for NAND/EQUIV)
    std::vector<unsigned int> input_literals;   ///< inputs of the gate (might be negated)
    std::vector<unsigned int> encoding_clauses; ///< IDs of the clauses encoding the gate
};


/**
 * \brief Represents a DQBF with a number of operations on it.
 */
class Formula
{
public:
    //@{
    /**\name Construction, destruction, and assigment
     */

    /**
     * \brief Constructs an empty formula.
     */
    Formula()                                = default;

    /**
     * \brief Creates a copy of an existing formula.
     */
    Formula(const Formula& other)            = default;

    /**
     * \brief Moves an existing formula into a new one.
     */
    Formula(Formula&& other)                 = default;

    /**
     * \brief Frees the memory occupied by a formula.
     */
    ~Formula() noexcept                      = default;

    /**
     * \brief Assigns a new formula to the current object.
     */
    Formula& operator=(const Formula& other) = default;

    /**
     * \brief Assigns a new formula to the current object (rvalue version)
     */
    Formula& operator=(Formula&& other)      = default;

    //@}

    //@{
    /**\name Formula input and output
     */
    void readFormula(std::istream& stream);
    void writeFormula(std::ostream& stream) const;
    //@}


    //@{
    /**\name Variable handling
     */
    void addUVar(unsigned int index);
    template <typename Container> void addEVar(unsigned int index, const Container& deps);
    unsigned int numVars()               const noexcept;
    unsigned int numUVars()              const noexcept;
    unsigned int numEVars()              const noexcept;
    bool isUniversal(unsigned int var)   const noexcept;
    bool isExistential(unsigned int var) const noexcept;
    unsigned int maxVarIndex()           const noexcept;
    unsigned int maxLitIndex()           const noexcept;
    void removeVar(unsigned int var);
    bool varDeleted(unsigned int var)    const noexcept;
    //@}

    //@{
    /**\name Clause handling
     */
    unsigned int numClauses()            const noexcept;
    unsigned int maxClauseIndex()        const noexcept;
    int addClause(const std::vector<unsigned int>& clause, bool needs_sorting = true);
    int addClause(std::vector<unsigned int>&& clause, bool needs_sorting = true);
    const std::vector<unsigned int>& getClause(unsigned int clause_nr) const noexcept;
    void removeClause(unsigned int c_nr);
    bool clauseDeleted(unsigned int c_nr) const noexcept;
    void writeClauses(std::ostream& stream) const;
    //@}


    //@{
    /**\name Dependency handling
     */
    unsigned int numDependencies() const noexcept;
    unsigned int numDependencies(unsigned int var) const noexcept;
    bool depends(unsigned int var1, unsigned int var2) const;
    void removeDependency(unsigned int var1, unsigned int var2);
    void clearDependencies(unsigned int var);
    void addDependency(unsigned int var1, unsigned int var2);
    const std::set<unsigned int>& getDependencies(unsigned int var) const;
    bool dependenciesSubset(unsigned int var1, unsigned int var2) const;
    bool applyDependencyScheme(DependencyScheme scheme, DependencyOperation operation, std::vector<unsigned int>* gate_outputs = nullptr);
    //@}


private:

    void reset();

    template <typename Function>
    void searchPath(
        const std::vector<unsigned int>& start_lits,
        Function forbidden,
        std::vector<bool>& seen) const;

    template <typename Function>
    void searchResolutionPath(
        const std::vector<unsigned int>& start_lits,
        Function forbidden,
        std::vector<bool>& seen) const;


    // Dependency schemes
    bool stdTriDep(bool resolution_paths, bool triangle);
    bool sstdQuadDep(bool resolution_paths, bool quadrangle);
    bool invStdTriDep(bool resolution_paths, bool triangle);
    bool invSstdQuadDep(bool resolution_paths, bool quadrangle);
    bool gateDependencies(DependencyOperation operation, std::vector<unsigned int>* gate_outputs = nullptr);


    // For gate detection
    bool checkGatePrecondition(unsigned int output_var, const std::vector<unsigned int>& clause) const;
    std::vector<Gate> determineGates();


    // Unit propagation
    void pushUnit(unsigned int lit, bool allow_universal = false);

    // Clause modification
    void addImplication(unsigned int lit1, unsigned int lit2);
    std::pair<bool, bool> hasImplicationTransitive(unsigned int start_lit, unsigned int target_var) const;
    bool hasImplication(unsigned int lit1, unsigned int lit2) const noexcept;
    void removeImplication(unsigned int lit1, unsigned int lit2);


    // DATA
    std::vector<VariableStatus> var_status;           ///< Status of the variables
    std::vector<ClauseStatus> clause_status;          ///< Status of the clauses
    std::vector<std::set<unsigned int>> prefix;       ///< The dependency sets of all variables (including universal ones)
    std::vector<std::vector<unsigned int>> clauses;   ///< the current list of clauses
    std::vector<std::vector<unsigned int>> occ_list;  ///< Specifies for each literal in which clauses it appears
    std::vector<std::set<unsigned int>> implications; ///< Stores the binary clauses redundantly as implications
    std::vector<unsigned int> deleted_clause_numbers; ///< Contains the IDs of those clauses which are currently deleted
    std::vector<unsigned int> unit_stack;             ///< Stack with unit literals; used by Formula::unitPropagation()
    std::vector<TruthValue> assignment;               ///< Assignment of unit literals; used by Formula::pushUnit() to detect conflicts
    mutable unsigned int num_u_vars = 0;              ///< Caches the number of universal variables
    mutable unsigned int num_e_vars = 0;              ///< Caches the number of existential variables

    // STATISTICS
    int stat_dep_schemes = 0;               ///< Change in the number of dependencies.
};

/**
 * \brief Output operator for DQBFs.
 *
 * Writes a DQBF in DQDIMACS format to the given stream.
 * \param stream the output stream the formula is written to
 * \param formula the formula that is to be written
 * \return the stream
 * \relates Formula
 */
std::ostream& operator<<(std::ostream& stream, const Formula& formula);


/**
 * \brief Input operator for DQBFs (in DQDIMACS format)
 *
 * Reads a formula in DQDIMACS format from an input stream.
 * \param stream the input stream the formula is read from
 * \param formula the data structure the read formula is stored in
 * \return the stream
 * \relates Formula
 */
std::istream& operator>>(std::istream& stream, Formula& formula);


/**
 * \brief Auxiliary class for printing clauses in DIMACS format.
 * \sa operator<<(ostream&, const PrintClause&)
 */
class PrintClause
{
public:
    /**
     * \brief Constructs a new PrintClause object for a given clause.
     */
    inline PrintClause(const std::vector<unsigned int>& _clause):
        clause(_clause) {}

    /**
     * \brief Print the given clause to the output stream.
     */
    inline void print(std::ostream& stream) const
    {
        for (const unsigned int lit: clause) stream << lit2dimacs(lit) << " ";
        stream << "0";
    }

private:
    const std::vector<unsigned int>& clause; ///< A reference to the clause that should be printed.
};

/**
 * \brief Used for printing clauses in DIMACS format.
 * \relates PrintClause
 * Example:<br>
 *      `std::vector<unsigned int> clause = ...;`<br>
 *      `std::cout << PrintClause(clause) << std::endl;`
 */
inline std::ostream& operator<<(std::ostream& stream, const PrintClause& clause)
{
    clause.print(stream);
    return stream;
}

#include "formula.ipp"

#endif
