/**
 * \file formula.ipp
 * \brief Inline functions for variable, clause, and dependency management
 * \author Ralf Wimmer
 * \date 12/2015-03/2016
 */


/**
 * \brief Returns the number of active variables.
 *
 * A variable is active if it is contained in at least one clause.
 * \sa Formula::numUVars()
 * \sa Formula::numEVars()
 */
inline unsigned int Formula::numVars() const noexcept
{
    return num_u_vars + num_e_vars;
}


/**
 * \brief Returns the number of active universal variables.
 *
 * A variable is active if it has not been deleted.
 * \sa Formula::numVars()
 * \sa Formula::numEVars()
 */
inline unsigned int Formula::numUVars() const noexcept
{
    return num_u_vars;
}


/**
 * \brief Returns the number of active existential variables.
 *
 * A variable is active if it has not been deleted.
 * \sa Formula::numVars()
 * \sa Formula::numUVars()
 */
inline unsigned int Formula::numEVars() const noexcept
{
    return num_e_vars;
}


/**
 * \brief Returns if a given variable is universal.
 *
 * A variable has one of three states: universal, existential, or deleted.
 * That means deleted variables are neither universal nor existential.
 * \return true iff the variable is universal.
 */
inline bool Formula::isUniversal(unsigned int var) const noexcept
{
    val_assert(var <= maxVarIndex());

    return var_status[var] == VariableStatus::UNIVERSAL;
}


/**
 * \brief Returns if a given variable is existential.
 *
 * A variable has one of three states: universal, existential, or deleted.
 * That means deleted variables are neither universal nor existential.
 * \return true iff the variable is existential.
 */
inline bool Formula::isExistential(unsigned int var) const noexcept
{
    val_assert(var <= maxVarIndex());

    return var_status[var] == VariableStatus::EXISTENTIAL;
}


/**
 * \brief Returns the maximal valid ID of a variable.
 *
 * Variable IDs always start at 1. Some of the IDs may be
 * inactive because the variables does not appear in the formula
 * anymore. Note that this function does not return the number
 * of active variables.
 * \sa Formula::numVars()
 * \sa Formula::maxLitIndex()
 */
inline unsigned int Formula::maxVarIndex() const noexcept
{ 
    return var_status.size() - 1;
}


/**
 * \brief Returns the maximal valid ID of a literal.
 *
 * Literal IDs always start at 2. Some of the IDs may be
 * inactive because the corresponding variables has been deleted.
 * Note that this function does not return the number
 * of active literals.
 * \sa Formula::numVars()
 * \sa Formula::maxVarIndex()
 */
inline unsigned int Formula::maxLitIndex() const noexcept
{
    return occ_list.size() - 1;;
}


/**
 * \brief Returns if a variable has been deleted.
 *
 * Variables are considered to be deleted if they do not
 * appear in the formula anymore.
 */
inline bool Formula::varDeleted(const unsigned int var) const noexcept
{
    val_assert(var <= maxVarIndex());

    return var_status[var] == VariableStatus::DELETED;
}


/**
 * \brief Removes a variable.
 *
 * \pre The variable may not occur in any clause anymore.
 */
inline void Formula::removeVar(const unsigned int var)
{
    val_assert(var <= maxVarIndex());
    val_assert(occ_list[var2lit(var, false)].empty());
    val_assert(occ_list[var2lit(var, true)].empty());

    if (varDeleted(var)) return;

    if (!prefix[var].empty()) clearDependencies(var);
    if (isUniversal(var)) --num_u_vars;
    else --num_e_vars;
    var_status[var] = VariableStatus::DELETED;
}


/**
 * \brief Creates a new existential variable.
 *
 * \param index the ID of the newly created variable
 * \param deps container with the dependency set of the created variable
 * \pre The created variable may not yet occur in the formula. All variables
 *      in the dependency set must be universal (and be created before).
 */
template <typename Container>
void Formula::addEVar(const unsigned int index, const Container& deps)
{
    assert(index > maxVarIndex() || varDeleted(index));

    if (index > maxVarIndex()) {
        var_status.resize(index + 1, VariableStatus::DELETED);
        prefix.resize(index + 1);
        occ_list.resize(2*index + 2);
        implications.resize(2*index + 2);
        assignment.resize(index + 1, TruthValue::UNKNOWN);
    }

    var_status[index] = VariableStatus::EXISTENTIAL;
    ++num_e_vars;
    auto pos = prefix[index].begin();
    for (unsigned int dep: deps) {
        assert(isUniversal(dep));
        pos = prefix[index].insert(pos, dep);
        prefix[dep].insert(index);
    }
}


/**
 * \brief Returns the number of clauses.
 */
inline unsigned int Formula::numClauses() const noexcept
{
    return clauses.size() - deleted_clause_numbers.size() + unit_stack.size();
}


/**
 * \brief Returns the maximum used clause ID.
 *
 * Clause IDs are numbered starting from 0. Some of the IDs may no
 * longer be used because clauses have been deleted. This can be checked
 * using Formula::clauseDeleted().
 */
inline unsigned int Formula::maxClauseIndex() const noexcept
{
    return clauses.size() - 1;
}


/**
 * \brief Returns a reference to the clause with the given ID.
 *
 * \pre The accessed clause may not be deleted.
 * \sa Formula::clauseDeleted()
 */
inline const std::vector<unsigned int>& Formula::getClause(unsigned int c_nr) const noexcept
{
    val_assert(c_nr <= maxClauseIndex());
    val_assert(!clauseDeleted(c_nr)); 

    return clauses[c_nr];
}


/**
 * \brief Checks if a clause has been deleted.
 */
inline bool Formula::clauseDeleted(unsigned int c_nr) const noexcept
{
    val_assert(c_nr <= maxClauseIndex());

    return clause_status[c_nr] == ClauseStatus::DELETED;
}


/**
 * \brief Returns the number of dependencies of a given variable.
 */
inline unsigned int Formula::numDependencies(unsigned int var) const noexcept
{
    val_assert(var <= maxVarIndex());

    return prefix[var].size();
}


/**
 * \brief Checks if one variable depends on another one.
 * \pre One of the variables needs to be universal, the other one existential.
 * \return true if the dependency between var1 and var2 exists.
 */
inline bool Formula::depends(unsigned int var1, unsigned int var2) const
{
    val_assert(var1 <= maxVarIndex() && var2 <= maxVarIndex()); 
    val_assert(isUniversal(var1) != isUniversal(var2));

    return prefix[var1].find(var2) != prefix[var1].cend();
}


/**
 * \brief Removes a dependency between two variables.
 * \pre One of the variables needs to be universal, the other one existential.
 */
inline void Formula::removeDependency(unsigned int var1, unsigned int var2)
{
    val_assert(var1 <= maxVarIndex() && var2 <= maxVarIndex());
    val_assert(isUniversal(var1) != isUniversal(var2));

    prefix[var1].erase(var2); prefix[var2].erase(var1);
}

/**
 * \brief Adds a dependency between two variables.
 * \pre One of the variables needs to be universal, the other one existential.
 */
inline void Formula::addDependency(unsigned int var1, unsigned int var2)
{
    val_assert(var1 <= maxVarIndex() && var2 <= maxVarIndex());
    val_assert(isUniversal(var1) != isUniversal(var2));

    prefix[var1].insert(var2); prefix[var2].insert(var1);
}


/**
 * \brief Returns a reference to the dependency set of a given variable.
 */
inline const std::set<unsigned int>& Formula::getDependencies(unsigned int var) const
{
    val_assert(var <= maxVarIndex()); 

    return prefix[var];
}


/**
 * \brief Adds an implication lit1 => lit2 to the implication data structure
 * \note The equivalent implication ~lit2 => ~lit1 is not added automatically.
 */
inline void Formula::addImplication(unsigned int lit1, unsigned int lit2)
{
    val_assert(lit1 < implications.size());
    val_assert(lit2 < implications.size());

    implications[lit1].insert(lit2);
}


/**
 * \brief Checks if the formula contains an implication lit1 => lit2.
 * \note This does not take transitivity into account.
 */
inline bool Formula::hasImplication(unsigned int lit1, unsigned int lit2) const noexcept
{
    val_assert(lit1 < implications.size());
    val_assert(lit2 < implications.size());

    return implications[lit1].find(lit2) != implications[lit1].cend();
}


/**
 * \brief Removes the implication lit1 => lit2 from the implication data structure.
 * \note This does neither remove the converse implication ~lit2 => ~lit1 nor the corresponding
 *       binary clause (~lit1, lit2).
 */
inline void Formula::removeImplication(unsigned int lit1, unsigned int lit2)
{
    val_assert(lit1 < implications.size());
    val_assert(lit2 < implications.size());

    implications[lit1].erase(lit2);
}
