#include <vector>
#include <stack>
#include <iostream>
#include "formula.hpp"

/**
 * \file formula_dep_schemes.cpp
 * \brief Implementation of dependency schemes for DQBF
 * \author Ralf Wimmer
 * \date 12/2015--03/2016
 */


/**
 * \brief Tries to find paths in the clause-literal incidence graph.
 *
 * \param start_lits the literals the paths may start at
 * \param forbidden a functor telling over which variable nodes the paths may not run
 * \param[out] seen a vector which stores the variables that can be reached by paths.
 * \sa stdTriDep(bool, bool)
 * \sa invStdTriDep(bool, bool)
 * \sa sstdQuadDep(bool, bool)
 * \sa invSstdQuadDep(bool, bool)
 */
template <typename Function>
void Formula::searchPath(
    const std::vector<unsigned int>& start_lits,
    Function forbidden,
    std::vector<bool>& seen) const
{
    val_assert(seen.size() > maxLitIndex());
    std::stack<unsigned int> open_lits;

    // Invariant:
    // - only literals are added to open_vars for which seen[lit] = false;
    // - all literals in open_vars are marked with seen[lit] = true
    // This ensures that no literal is added to open_vars twice.
    for (unsigned int lit: start_lits) {
        if (!seen[lit]) open_lits.push(lit);
        seen[lit] = true;
    }

    while (!open_lits.empty()) {
        const auto current_lit = open_lits.top();
        open_lits.pop();
        for (unsigned int c_nr: occ_list[current_lit]) {
            val_assert(!clauseDeleted(c_nr));

            for (unsigned int next_lit: clauses[c_nr]) {
                if (!seen[next_lit] && !forbidden(next_lit)) open_lits.push(next_lit);
                if (!seen[negate(next_lit)] && !forbidden(negate(next_lit))) open_lits.push(negate(next_lit));

                seen[next_lit] = true;
                seen[negate(next_lit)] = true;
            }
        }
    }
}


/**
 * \brief Tries to find resolution paths in the clause-literal incidence graph.
 *
 * A resolution path connects clauses in which the connecting variable
 * appears in different polarities. Additionally, the same variable may not
 * be used twice in sequence along the path.
 * \param start_lits the literals the resolution paths may start at
 * \param forbidden a functor telling over which variable nodes the paths may not run
 * \param[out] seen a vector which stores the variables that can be reached by paths.
 * \sa stdTriDep(bool, bool)
 * \sa invStdTriDep(bool, bool)
 * \sa sstdQuadDep(bool, bool)
 * \sa invSstdQuadDep(bool, bool)
 */
template <typename Function>
void Formula::searchResolutionPath(
    const std::vector<unsigned int>& start_lits,
    Function forbidden,
    std::vector<bool>& seen) const
{
    val_assert(seen.size() > maxLitIndex());
    std::stack<unsigned int> open_lits;

    // Invariant:
    // - only literals are added to open_vars for which seen[lit] = false;
    // - all literals in open_vars are marked with seen[lit] = true
    // This ensures that no literal is added to open_vars twice.
    for (unsigned int lit: start_lits) {
        for (unsigned int c_nr: occ_list[lit]) {
            val_assert(!clauseDeleted(c_nr));

            for (unsigned int next_lit: clauses[c_nr]) {
                if (!forbidden(next_lit) && !seen[next_lit]) {
                    open_lits.push(next_lit);
                }
                seen[next_lit] = true;
            }
        }
    }

    while (!open_lits.empty()) {
        const auto current_lit = negate(open_lits.top());
        open_lits.pop();
        for (unsigned int c_nr: occ_list[current_lit]) {
            val_assert(!clauseDeleted(c_nr));

            for (unsigned int next_lit: clauses[c_nr]) {
                if (!seen[next_lit] && !forbidden(next_lit)) {
                    open_lits.push(next_lit);
                }
                seen[next_lit] = true;
            }
        }
    }
}


/**
 * \brief Implementation of the standard and reflexive triangle dependency scheme for addinging pseudo-dependencies
 *
 * The dependency scheme can either use normal
 * or resolution paths. Identified pseudo-dependencies are
 * added to the dependency sets.
 * \param resolution_paths if true, use resolution paths
 * \param triangle if true, use the reflexive triangle dependency scheme
 * \return true if the formula was modified
 * \sa stdTriDep(bool, bool)
 * \sa sstdQuadDep(bool, bool)
 * \sa invSstdQuadDep(bool, bool)
 */
bool Formula::invStdTriDep(bool resolution_paths, bool triangle)
{
    int count = 0;
    std::vector<bool> seen(maxLitIndex() + 1, false);

    for (unsigned int u_var = 1; u_var <= maxVarIndex(); ++u_var) {
        if (isExistential(u_var)) continue;

        const auto forbidden = [this, u_var](unsigned int lit) -> bool
        {
            return this->isUniversal(lit2var(lit)) || !this->depends(lit2var(lit), u_var);
        };

        seen.assign(seen.size(), false);
        if (resolution_paths)
            searchResolutionPath({var2lit(u_var, true), var2lit(u_var, false)}, forbidden, seen);
        else
            searchPath({var2lit(u_var, true), var2lit(u_var, false)}, forbidden, seen);

        for (unsigned int e_var = 1; e_var <= maxVarIndex(); ++e_var) {
            if (isUniversal(e_var) || depends(e_var, u_var)) continue;

            const unsigned int ep = var2lit(e_var, false);
            const unsigned int en = var2lit(e_var, true);

            if ( (triangle && (!seen[ep] || !seen[en]))
                  || (!triangle && !seen[ep] && !seen[en]) ) {
                addDependency(u_var, e_var);
                ++count;
            }
        }
    }

    stat_dep_schemes += count;
    return count > 0;
}


/**
 * \brief Implementation of the strict standard and reflexive quadrangle dependency scheme for addinging pseudo-dependencies
 *
 * The dependency scheme can either use normal
 * or resolution paths. Identified pseudo-dependencies are
 * added to the dependency sets.
 * \param resolution_paths if true, use resolution paths
 * \param quadrangle if true, use the reflexive quadrangle dependency scheme
 * \return true if the formula was modified
 * \sa stdTriDep(bool, bool)
 * \sa invStdTriDep(bool, bool)
 * \sa sstdQuadDep(bool, bool)
 */
bool Formula::invSstdQuadDep(bool resolution_paths, bool quadrangle)
{
    int count = 0;
    std::vector<bool> seen_pos(maxLitIndex() + 1, false);
    std::vector<bool> seen_neg(maxLitIndex() + 1, false);

    for (unsigned int u_var = 1; u_var <= maxVarIndex(); ++u_var) {
        if (isExistential(u_var)) continue;

        const auto forbidden = [this, u_var](unsigned int lit) -> bool
        {
            return this->isUniversal(lit2var(lit)) || !this->depends(lit2var(lit), u_var);
        };

        seen_pos.assign(seen_pos.size(), false);
        seen_neg.assign(seen_neg.size(), false);

        if (resolution_paths) {
            searchResolutionPath({var2lit(u_var,false)}, forbidden, seen_pos);
            searchResolutionPath({var2lit(u_var, true)}, forbidden, seen_neg);
        } else {
            searchPath({var2lit(u_var,false)}, forbidden, seen_pos);
            searchPath({var2lit(u_var, true)}, forbidden, seen_neg);
        }

        for (unsigned int e_var = 1; e_var <= maxVarIndex(); ++e_var) {
            if (isUniversal(e_var) || depends(e_var, u_var)) continue;

            const unsigned int ep = var2lit(e_var, false);
            const unsigned int en = var2lit(e_var, true);

            if (quadrangle) {
                if ( (!seen_pos[ep] || !seen_neg[ep]) && (!seen_neg[ep] || !seen_pos[en])) {
                    addDependency(u_var, e_var);
                    ++count;
                }
            } else {
                if ( (!seen_pos[ep] && !seen_pos[en]) || (!seen_neg[ep] && !seen_neg[en])) {
                    addDependency(u_var, e_var);
                    ++count;
                }
            }
        }
    }

    stat_dep_schemes += count;
    return count > 0;
}


/**
 * \brief Implementation of the standard and reflexive triangle dependency scheme for removing pseudo-dependencies
 *
 * The dependency scheme can either use normal
 * or resolution paths. Identified pseudo-dependencies are
 * removed from the dependency sets.
 * \param resolution_paths if true, use resolution paths
 * \param triangle if true use the reflexive triangle instead of the standard dependency scheme
 * \return true if the formula was modified
 * \sa invStdTriDep(bool, bool)
 * \sa sstdQuadDep(bool, bool)
 * \sa invSstdQuadDep(bool, bool)
 */
bool Formula::stdTriDep(bool resolution_paths, bool triangle)
{
    int count = 0;
    std::vector<bool> seen(maxLitIndex() + 1, false);

    for (unsigned int u_var = 1; u_var <= maxVarIndex(); ++u_var) {
        if (!isUniversal(u_var)) continue;

        const auto forbidden = [this, u_var](unsigned int lit) -> bool
        {
            return isUniversal(lit2var(lit)) || !depends(lit2var(lit), u_var);
        };

        // Peforming depth-first search starting at 'u_var'.
        seen.assign(seen.size(), false);
        if (resolution_paths)
            searchResolutionPath({var2lit(u_var, true), var2lit(u_var,false)}, forbidden, seen);
        else {
            searchPath({var2lit(u_var, true), var2lit(u_var,false)}, forbidden, seen);
        }

        for (unsigned int e_var = 1; e_var <= maxVarIndex(); ++e_var) {
            if (!isExistential(e_var) || !depends(e_var, u_var)) continue;

            const unsigned int ep = var2lit(e_var, false);
            const unsigned int en = var2lit(e_var, true);

            if ( (triangle && (!seen[en] || !seen[ep]))
                 || (!triangle && !seen[en] && !seen[ep]) ) {
                removeDependency(e_var, u_var);
                ++count;
            }
        }
    } // end for u_var

    stat_dep_schemes -= count;
    return count > 0;
}


/**
 * \brief Implementation of the strict standard and reflexive quadrangle dependency scheme for removing pseudo-dependencies
 *
 * The dependency scheme can either use normal
 * or resolution paths. Identified pseudo-dependencies are
 * removed from the dependency sets.
 * \param resolution_paths if true, use resolution paths
 * \param quadrangle if true, reflexive quadrangle dependencies are computed
 * \return true if the formula was modified
 * \sa stdTriDep(bool, bool)
 * \sa invStdTriDep(bool, bool)
 * \sa invSstdQuadDep(bool, bool)
 */
bool Formula::sstdQuadDep(bool resolution_paths, bool quadrangle)
{
    int count = 0;
    std::vector<bool> seen_pos(maxLitIndex() + 1, false);
    std::vector<bool> seen_neg(maxLitIndex() + 1, false);

    for (unsigned int u_var = 1; u_var <= maxVarIndex(); ++u_var) {
        if (!isUniversal(u_var)) continue;

        const auto forbidden = [this, u_var](unsigned int lit) -> bool
        {
            return isUniversal(lit2var(lit)) || !depends(lit2var(lit), u_var);
        };

        seen_pos.assign(seen_pos.size(), false);
        seen_neg.assign(seen_neg.size(), false);

        if (resolution_paths) {
            // Peforming depth-first search starting at 'u_var'.
            searchResolutionPath({var2lit(u_var, false)}, forbidden, seen_pos);

            // Peforming depth-first search starting at '!u_var'.
            searchResolutionPath({var2lit(u_var, true)}, forbidden, seen_neg);
        } else {
            // Peforming depth-first search starting at 'u_var'.
            searchPath({var2lit(u_var, false)}, forbidden, seen_pos);

            // Peforming depth-first search starting at '!u_var'.
            searchPath({var2lit(u_var, true)}, forbidden, seen_neg);
        }

        for (unsigned int e_var = 1; e_var <= maxVarIndex(); ++e_var) {
            if (!isExistential(e_var) || !depends(e_var, u_var)) continue;

            const unsigned int ep = var2lit(e_var, false);
            const unsigned int en = var2lit(e_var, true);

            if (quadrangle) {
                if ( (!seen_pos[ep] || !seen_neg[ep]) || (!seen_neg[ep] && !seen_pos[en]) ) {
                    removeDependency(e_var, u_var);
                    ++count;
                }
            } else {
                if ( (!seen_pos[en] && !seen_pos[ep]) || (!seen_neg[en] && !seen_neg[ep]) ) {
                    removeDependency(e_var, u_var);
                    ++count;
                }
            }
        }
    } // end for u_var

    stat_dep_schemes -= count;
    return count > 0;
}


/**
 * \brief Uses gate detection to identify pseudo-dependencies.
 * \param operation specifies whether pseudo-dependencies are added, removed, or only identified.
 * \param gate_outputs if not nullptr, it contains afterwards the gate output variables
 *        in topological order.
 * \return true if the formula was modified.
 * \todo Update statistics on the number of modified dependencies.
 */
bool Formula::gateDependencies(DependencyOperation operation, std::vector<unsigned int>* gate_outputs)
{
    int count = 0;

    std::vector<Gate> gates = determineGates();
    if (gate_outputs != nullptr) {
        for (const auto& g: gates) {
            gate_outputs->push_back(lit2var(g.output_literal));
        }
    }

    if (operation == DependencyOperation::ADD) {
        // We can make all gate outputs depend on all universal variables
        // as they are implied by the gate inputs
        std::set<unsigned int> univ_vars;
        for (unsigned int var = 1; var <= maxVarIndex(); ++var) {
            if (isUniversal(var)) univ_vars.insert(var);
        }

        for (const auto& g: gates) {
            const unsigned int output_var = lit2var(g.output_literal);
            if (prefix[output_var].size() < univ_vars.size()) {
                count += univ_vars.size() - prefix[output_var].size();
                prefix[output_var] = univ_vars;
                for (unsigned int dep: univ_vars) prefix[dep].insert(output_var);
            }
        }
    } else if (operation == DependencyOperation::REMOVE) {
        for (const auto& g: gates) {
            const unsigned int outp_var = lit2var(g.output_literal);
            std::set<unsigned int> new_deps;
            for (unsigned int inp_lit: g.input_literals) {
                const unsigned int inp_var = lit2var(inp_lit);
                if (isUniversal(inp_var)) new_deps.insert(inp_var);
                else new_deps.insert(prefix[inp_var].cbegin(), prefix[inp_var].cend());
            }
            prefix[outp_var].swap(new_deps);
            for (unsigned int dep: new_deps) {
                if (prefix[outp_var].find(dep) == prefix[outp_var].cend()) {
                    prefix[dep].erase(outp_var);
                    --count;
                }
            }
        }
    }

    stat_dep_schemes += count;
    return count != 0;
}


/**
 * \brief Applies a dependency scheme to the formula to manipulate the dependency sets.
 * \param scheme specifies which dependency scheme is applied
 * \param operation determines whether dependencies should be added or removed
 * \param gate_outputs if scheme is DependencyScheme::GATE, this parameter is filled with the gate ouputs.
 * \return true if the formula was modified.
 */
bool Formula::applyDependencyScheme(
    DependencyScheme scheme,
    DependencyOperation operation,
    std::vector<unsigned int>* gate_outputs)
{
    switch(scheme) {
    case DependencyScheme::TRIVIAL:
        return false;
    case DependencyScheme::STANDARD:
        if (operation == DependencyOperation::REMOVE) return stdTriDep(false, false);
        else if (operation == DependencyOperation::ADD) return invStdTriDep(false, false);
        break;
    case DependencyScheme::STRICT_STANDARD:
        if (operation == DependencyOperation::REMOVE) return sstdQuadDep(false, false);
        else if (operation == DependencyOperation::ADD) return invSstdQuadDep(false, false);
        break;
    case DependencyScheme::REF_TRIANGLE:
        if (operation == DependencyOperation::REMOVE) return stdTriDep(false, true);
        else if (operation == DependencyOperation::ADD) return invStdTriDep(false, true);
        break;
    case DependencyScheme::REF_QUADRANGLE:
        if (operation == DependencyOperation::REMOVE) return sstdQuadDep(false, true);
        else if (operation == DependencyOperation::ADD) return invSstdQuadDep(false, true);
        break;
    case DependencyScheme::RP_STANDARD:
        if (operation == DependencyOperation::REMOVE) return stdTriDep(true, false);
        else if (operation == DependencyOperation::ADD) return invStdTriDep(true, false);
        break;
    case DependencyScheme::RP_STRICT_STANDARD:
        if (operation == DependencyOperation::REMOVE) return sstdQuadDep(true, false);
        else if (operation == DependencyOperation::ADD) return invSstdQuadDep(true, false);
        break;
    case DependencyScheme::RP_REF_TRIANGLE:
        if (operation == DependencyOperation::REMOVE) return stdTriDep(true, true);
        else if (operation == DependencyOperation::ADD) return invStdTriDep(true, true);
        break;
    case DependencyScheme::RP_REF_QUADRANGLE:
        if (operation == DependencyOperation::REMOVE) return sstdQuadDep(true, true);
        else if (operation == DependencyOperation::ADD) return invSstdQuadDep(true, true);
        break;
    case DependencyScheme::GATE:
        return gateDependencies(operation, gate_outputs);
        break;
    default:
        std::cerr << "[ERROR] Invalid dependency scheme." << std::endl;
        return false;
        break;
    }
    return false;
}
