#include <vector>
#include <set>
#include <map>
#include <algorithm>
#include <stack>
#include <functional>
#include <iterator>
#include "formula.hpp"

/**
 * \file formula_gates.cpp
 * \brief Implementation of gate detection
 * \author Ralf Wimmer
 * \date 01/2016
 */


/**
 * \brief Checks if a given variable is suited to be a gate output.
 *
 * A variable can only be a gate output if it is existential and depends
 * at least on the variable the gate inputs depend on. This function does
 * not check if all necessary clauses are present.
 * \param output_var the variable that is assumed to be a gate output
 * \param clause a clause that contains the gate inputs and the gate output
 * \return true if the described conditions are satisfied
 * \sa Formula::determineGates()
 */
bool Formula::checkGatePrecondition(unsigned int output_var, const std::vector<unsigned int>& clause) const
{
    if (isUniversal(output_var)) return false;
    for (auto lit: clause) {
        const auto current_var = lit2var(lit);
        if (current_var == output_var) continue;
        if (!dependenciesSubset(current_var, output_var)) return false;
    }
    return true;
}

/**
 * \brief Adds the dependencies of a single gate to gateDep
 *
 * This function is used by the gate detection for analyzing
 * whether a gate would create a cycle in the circuit.
 * \sa cyclicDep
 * \sa Formula::detectGates()
 */
static void addDep(std::vector<std::vector<unsigned int>>& gateDep, unsigned int base_var, const std::vector<unsigned int>& parents)
{
    if (gateDep[base_var].empty()) gateDep[base_var] = parents;
    else std::copy(parents.cbegin(), parents.cend(), std::back_inserter(gateDep[base_var]));
}


/**
 * \brief Returns true if adding another gate would create a cyclic dependency.
 *
 * This function is used by the gate detection for analyzing whether a gate
 * would create a cycle in the circuit.
 * \sa addDep
 * \sa Formula::detectGates
 */
static bool cyclicDep(std::vector<std::vector<unsigned int>>& gateDep, unsigned int base_var, const std::vector<unsigned int>& parents)
{
    std::stack<unsigned int> pending;
    std::vector<bool> processed(gateDep.size(), false);

    for (const unsigned int p: parents) {
        val_assert(p != base_var);
        pending.push(p);
    }

    while (!pending.empty()) {
        const unsigned int pt = pending.top();
        pending.pop();

        if (!processed[pt]) {
            const auto& dep = gateDep[pt];
            processed[pt] = true;

            for (const unsigned int p: dep) {
                if (p == base_var) return true;
                else if (!processed[p]) pending.push(p);
            }
        }
    }
    return false;
}


/**
 * \brief Checks if two clauses contain the same variables (with different polarities).
 * \pre Requires c1 and c2 to be sorted.
 * \pre Clauses must not be tautological or contain duplicate literals.
 */
static bool sameVars(const std::vector<unsigned int>& c1, const std::vector<unsigned int>& c2)
{
    if (c1.size() != c2.size()) return false;
    for (unsigned int i = 0; i < c1.size(); ++i) {
        if (lit2var(c1[i]) != lit2var(c2[i])) return false;
    }
    return true;
}

/**
 * \brief Counts the number of positive literals in a clause.
 */
static unsigned int posCount(const std::vector<unsigned int>& clause)
{
    return std::count_if(clause.cbegin(), clause.cend(), isPositive);
}


/**
 * \brief Performs gate detection on a quantified formula in CNF.
 *
 * We can detect 2-input XOR and multi-input AND-gates with arbitrarily
 * negated inputs and outputs. The gates are returned in topological order
 * (from inputs to outputs).
 * \sa Formula::substituteGate(const Gate&)
 */
std::vector<Gate> Formula::determineGates()
{
    std::vector<Gate> gates;

    std::map<std::pair<unsigned int, unsigned int>, unsigned int> bin2c_nr;
    for (unsigned int c_nr = 0; c_nr < clauses.size(); ++c_nr) {
        if (!clauseDeleted(c_nr) && clauses[c_nr].size() == 2) {
            bin2c_nr[std::make_pair(clauses[c_nr][0], clauses[c_nr][1])] = c_nr;
        }
    }

    std::vector<bool> clause_used(clauses.size(), false);
    std::vector<bool> is_output(maxVarIndex() + 1, false);
    std::vector<bool> is_input(maxVarIndex() + 1, false);
    std::vector<std::vector<unsigned int>> gateDep(maxVarIndex() + 1);

    // Looking for (multi-)AND-gates with arbitrarily negated inputs and output
    for (unsigned int c_nr = 0; c_nr < clauses.size(); ++c_nr) {
        if (clauseDeleted(c_nr) || clauses[c_nr].size() <= 2 || clause_used[c_nr]) continue;

        for (unsigned int l_nr = 0; l_nr < clauses[c_nr].size(); ++l_nr)
        {
            const unsigned int base_lit = clauses[c_nr][l_nr];
            const unsigned int base_var = lit2var(base_lit);
            if (is_output[base_var]) continue;

            std::vector<unsigned int> parents;
            parents.reserve(clauses[c_nr].size() - 1);
            bool has_imp = true;
            for (unsigned int lit: clauses[c_nr]) {
                if (lit == base_lit) continue;
                if (!hasImplication(base_lit, negate(lit))) { has_imp = false; break; }
                parents.push_back(lit2var(lit));
            }
            if (!has_imp) continue;
            if (!checkGatePrecondition(base_var, clauses[c_nr])) continue;

            // Check for cyclic dependencies
            if (is_input[base_var] && cyclicDep(gateDep, base_var, parents)) continue;

            addDep(gateDep, base_var, parents);
            clause_used[c_nr] = true;
            is_output[base_var] = true;
            for (unsigned int p: parents) is_input[p] = true;

            // We have found an AND-gate
            Gate g;
            g.type = GateType::AND_GATE;
            g.output_literal = base_lit;
            g.encoding_clauses.reserve(clauses[c_nr].size());
            g.input_literals.reserve(clauses[c_nr].size() - 1);
            g.encoding_clauses.push_back(c_nr);
            clause_status[c_nr] = ClauseStatus::MANDATORY;
            for (unsigned int lit: clauses[c_nr]) {
                if (lit == base_lit) continue;
                g.input_literals.push_back(negate(lit));
                const auto inp_c_nr = (negate(base_lit) < negate(lit) ?
                    bin2c_nr.find(std::make_pair(negate(base_lit), negate(lit))) :
                    bin2c_nr.find(std::make_pair(negate(lit), negate(base_lit))));
                val_assert(inp_c_nr != bin2c_nr.cend());
                g.encoding_clauses.push_back(inp_c_nr->second);
                clause_status[inp_c_nr->second] = ClauseStatus::MANDATORY;
            }

            gates.push_back(std::move(g));
            break;
        }
    }


    // Looking for XOR-gates with arbitrarily negated inputs and output
    for (unsigned int base_var = 1; base_var <= maxVarIndex(); ++base_var)
    {
        if (isUniversal(base_var)) continue;
        if (is_output[base_var]) continue;
        if (occ_list[var2lit(base_var, false)].size() < 2 || occ_list[var2lit(base_var, true)].size() < 2) continue;

        const unsigned int base_lit_pos = var2lit(base_var, false);
        const unsigned int base_lit_neg = var2lit(base_var, true);
        bool found_gate = false;

        for (unsigned int p1 = 0; p1 < occ_list[base_lit_pos].size(); ++p1) {
            const auto i1 = occ_list[base_lit_pos][p1];
            val_assert(!clauseDeleted(i1));
            const auto& c1 = clauses[i1];
            if (c1.size() != 3) continue;
            if (clause_used[i1]) continue;
            if (!checkGatePrecondition(base_var, c1)) continue;

            for (unsigned int p2 = p1 + 1; p2 < occ_list[base_lit_pos].size(); ++p2) {
                const unsigned int i2 = occ_list[base_lit_pos][p2];
                val_assert(!clauseDeleted(i2));
                const auto& c2 = clauses[i2];
                if (clause_used[i2]) continue;
                if (!sameVars(c1,c2)) continue;

                for (unsigned int p3 = 0; p3 < occ_list[base_lit_neg].size(); ++p3) {
                    const unsigned int i3 = occ_list[base_lit_neg][p3];
                    val_assert(!clauseDeleted(i3));
                    const auto& c3 = clauses[i3];
                    if (clause_used[i3]) continue;
                    if (!sameVars(c1,c3)) continue;

                    for (unsigned int p4 = p3 + 1;  p4 < occ_list[base_lit_neg].size(); ++p4) {
                        const unsigned int i4 = occ_list[base_lit_neg][p4];
                        val_assert(!clauseDeleted(i4));
                        const auto& c4 = clauses[i4];
                        if (clause_used[i4]) continue;
                        if (!sameVars(c1,c4)) continue;

                        if (c1 == c2 || c1 == c3 || c1 == c4 || c2 == c3 || c2 == c4 || c3 == c4) continue;

                        const auto pc1 = posCount(c1);
                        const auto pc2 = posCount(c2);
                        const auto pc3 = posCount(c3);
                        const auto pc4 = posCount(c4);

                        if( !( (  ( pc1 & 1 ) &&  ( pc2 & 1 ) &&  ( pc3 & 1 ) &&  ( pc4 & 1 ) ) ||
                               ( !( pc1 & 1 ) && !( pc2 & 1 ) && !( pc3 & 1 ) && !( pc4 & 1 ) ) ) )
                        {
                            // mixed polarities -> skipping
                            continue;
                        }


                        std::vector<unsigned int> parents(2);
                        bool base_found = false;
                        if (c1[0] == base_lit_pos) {
                            parents[0] = lit2var(c1[1]);
                            parents[1] = lit2var(c1[2]);
                            if (!is_input[base_var] || !cyclicDep(gateDep, base_var, parents)) {
                                addDep(gateDep, base_var, parents);
                                base_found = true;
                            }
                        } else if (c1[1] == base_lit_pos) {
                            parents[0] = lit2var(c1[0]);
                            parents[1] = lit2var(c1[2]);
                            if (!is_input[base_var] || !cyclicDep(gateDep, base_var, parents)) {
                                addDep(gateDep, base_var, parents);
                                base_found = true;
                            }
                        } else if (c1[2] == base_lit_pos) {
                            parents[0] = lit2var(c1[0]);
                            parents[1] = lit2var(c1[1]);
                            if (!is_input[base_var] || !cyclicDep(gateDep, base_var, parents)) {
                                addDep(gateDep, base_var, parents);
                                base_found = true;
                            }
                        }

                        if (!base_found) continue;

                        // Now 'base_var' is the output of an XOR gate.
                        Gate g;
                        g.type = GateType::XOR_GATE;
                        g.output_literal = var2lit(base_var, (pc1 & 1) == 1);
                        g.input_literals = {var2lit(parents[0]), var2lit(parents[1])};
                        g.encoding_clauses = { i1, i2, i3, i4 };
                        clause_status[i1] = ClauseStatus::MANDATORY;
                        clause_status[i2] = ClauseStatus::MANDATORY;
                        clause_status[i3] = ClauseStatus::MANDATORY;
                        clause_status[i4] = ClauseStatus::MANDATORY;
                        gates.push_back(std::move(g));

                        is_input[parents[0]] = true;
                        is_input[parents[1]] = true;
                        is_output[base_var] = true;
                        clause_used[i1] = true;
                        clause_used[i2] = true;
                        clause_used[i3] = true;
                        clause_used[i4] = true;
                        found_gate = true;
                        break;
                    } // for c4
                    if (found_gate) break;
                } // for c3
                if (found_gate) break;
            } // for c2
            if (found_gate) break;
        } // for c1
    } // for base_var


    // Create a topological ordering of the variables in gateDep.
    std::vector<int> which_gate(maxVarIndex() + 1, -1);
    for (unsigned int i = 0; i < gates.size(); ++i) which_gate[lit2var(gates[i].output_literal)] = i;

    std::vector<bool> processed(maxVarIndex() + 1, false);
    for (unsigned int var = 1; var <= maxVarIndex(); ++var) {
        processed[var] = !is_output[var];
    }

    // dfs is a recursive function which performs depth-first search on the
    // graph defined by the gate dependencies.
    std::vector<unsigned int> order;
    order.reserve(gates.size());
    std::function<void(int)> dfs = [this, &gateDep, &processed, &dfs, &order](unsigned int var) -> void
        {
            for (unsigned int dep: gateDep[var]) {
                if (!processed[dep]) dfs(dep);
            }
            processed[var] = true;
            order.push_back(var);
            }; // end lambda function

    for (unsigned int var = 1; var <= maxVarIndex(); ++var) {
        if (!processed[var] && !is_input[var]) dfs(var);
    }

    std::vector<Gate> result;
    result.reserve(gates.size());
    for (unsigned int var: order) {
        val_assert(which_gate[var] != -1);
        result.push_back(std::move(gates[which_gate[var]]));
    }

    return result;
}
