#include <string>
#include <sstream>
#include <set>
#include <vector>
#include <algorithm>
#include <iostream>
#include <iterator>
#include "aux.hpp"
#include "formula.hpp"

/**
 * \file formula_inout.cpp
 * \brief Implementation of methods and functions for reading and writing formulas.
 * \author Ralf Wimmer
 * \date 01/2016
 */


/**
 * \brief Reads a formula in DQDIMACS format from an input stream
 *
 * If the current formula is not empty, it is cleared first by calling Formula::reset().
 */
void Formula::readFormula(std::istream& stream)
{
    reset();

    std::string line;
    unsigned int num_vars = 0;
    unsigned int num_clauses = 0;
    std::set<unsigned int> universal_vars;

    while (std::getline(stream, line)) {
        std::istringstream stream(line);

        char next_char = stream.peek();
        if (next_char == 'p') {
            // header
            stream >> next_char; // eat up the character 'p'
            std::string s;
            stream >> s; // cnf
            stream >> num_vars;
            stream >> num_clauses;
            var_status.resize(num_vars + 1, VariableStatus::DELETED);
            clause_status.resize(num_clauses + 1, ClauseStatus::DELETED);
            clauses.reserve(num_clauses + 1);
            prefix.resize(num_vars + 1, std::set<unsigned int>());
            occ_list.resize(2*num_vars + 2);
            implications.resize(2*num_vars + 2);
            assignment.assign(num_vars + 1, TruthValue::UNKNOWN);
        } else if (next_char == 'a') {
            // list of universally quantified variables
            stream >> next_char; // eat up the character 'a'
            while (stream) {
                unsigned int var = 0;
                stream >> var;
                if (var == 0) break;
                val_assert(var <= maxVarIndex());
                var_status[var] = VariableStatus::UNIVERSAL;
                universal_vars.insert(var);
            }
        } else if (next_char == 'e') {
            // existential variable depending on all universal vars 
            // declared so far.
            stream >> next_char; // eat up the character 'e'
            unsigned int exist_var = 0;
            while (stream) {
                stream >> exist_var;
                if (exist_var == 0) break;
                val_assert(exist_var <= maxVarIndex());
                var_status[exist_var] = VariableStatus::EXISTENTIAL;
                prefix[exist_var] = universal_vars;
                for (unsigned int univ_var: universal_vars) prefix[univ_var].insert(exist_var);
            }
        } else if (next_char == 'd') {
            // existential variable with dependencies
            stream >> next_char; // eat up the character 'd'
            unsigned int exist_var = 0;
            unsigned int all_var = 0;

            stream >> exist_var;
            val_assert(exist_var <= maxVarIndex());
            var_status[exist_var] = VariableStatus::EXISTENTIAL;

            while (stream) {
                stream >> all_var;
                if (all_var == 0) break;
                prefix[exist_var].insert(all_var);
                prefix[all_var].insert(exist_var);
            }
        } else if (next_char == 'c') continue; // comment line
        else if (next_char == '-' || (next_char >= '0' && next_char <= '9')) {
            // clause
            std::vector<unsigned int> clause;
            int lit = 0;
            while (stream) {
                stream >> lit;
                if (lit == 0) break;
                const unsigned int var = abs(lit);
                clause.push_back(var2lit(var, lit < 0));
            }
            addClause(std::move(clause));
        }
    }
    num_u_vars = universal_vars.size();
    num_e_vars = num_vars - num_u_vars;
}

/**
 * \brief Writes the current formula in DQDIMACS format to the given output stream.
 */
void Formula::writeFormula(std::ostream& stream) const
{
    if (!stream) {
        std::cerr << "[ERROR] Could not write to output stream!" << std::endl;
        std::exit(-1);
    }

    // print the header
    stream << "p cnf " << maxVarIndex() << " " << numClauses() << std::endl;

    stream << "c deleted variables:\n";
    stream << "e ";
    for (unsigned int var = 1; var <= maxVarIndex(); ++var) {
        if (varDeleted(var)) stream << var << " ";
    }
    stream << "0" << std::endl;

    // print universal variables
    unsigned int num_univ = 0;
    stream << "a ";
    for (unsigned int var = 1; var <= maxVarIndex(); ++var) {
        if (isUniversal(var)) {
            stream << var << " ";
            ++num_univ;
        }
    }
    stream << "0" << std::endl;

    // print existential variables which do not depend on all universals
    bool var_skipped = false;
    for (unsigned int var = 1; var <= maxVarIndex(); ++var) {
        if (!isExistential(var)) continue;
        if (prefix[var].size() == num_univ) {
            // Skip variables which depend on all universals.
            var_skipped = true;
            continue;
        }
        stream << "d " << var << " ";
        std::copy(prefix[var].cbegin(), prefix[var].cend(), std::ostream_iterator<unsigned int>(stream, " "));
        stream << "0" << std::endl;
    }

    // print existential variable which depend on all universals.
    if (var_skipped) {
        stream << "e ";
        for (unsigned int var = 1; var <= maxVarIndex(); ++var) {
            if (!isExistential(var)) continue;
            if (prefix[var].size() == num_univ) stream << var << " ";
        }
        stream << "0" << std::endl;
    }

    writeClauses(stream);
}

/**
 * \brief Print all clauses in DIMACS format to the given stream.
 */
void Formula::writeClauses(std::ostream& stream) const
{
    // Print the clauses.
    for (unsigned int c_nr = 0; c_nr < clauses.size(); ++c_nr) {
        if (clauseDeleted(c_nr)) continue;
        stream << PrintClause(clauses[c_nr]) << std::endl;
    }

    for (unsigned int lit: unit_stack) {
        stream << lit2dimacs(lit) << " 0" << std::endl;
    }
}


std::ostream& operator<<(std::ostream& stream, const Formula& formula)
{
    formula.writeFormula(stream);
    return stream;
}


std::istream& operator>>(std::istream& stream, Formula& formula)
{
    formula.readFormula(stream);
    return stream;
}
