// $Id: main.cpp 776 2016-04-19 09:19:36Z wimmer $

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <vector>
#include <set>
#include <iterator>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/program_options.hpp>
#include "aux.hpp"
#include "formula.hpp"

int main(int argc, char** argv)
{
    std::cout << "\n\nThis is HQS Dependency Analyzer, compiled on " << __DATE__ << " at " << __TIME__ << std::endl << std::endl;

    std::string in_name("");
    std::string out_name("");

    // Declaration of parameters:
    boost::program_options::options_description public_options("Options");
    public_options.add_options()
    ("help,h", "Show available options")
    ("std", "Apply standard dependency scheme")
    ("sstd", "Apply strict standard dependency scheme")
    ("tri", "Apply reflexive triangle dependency scheme")
    ("quad", "Apply reflexive quadrangle dependency scheme")
    ("gate", "Use gate detection")
    ("rpath", "Use resolution paths for the dependency scheme")
    ("add,a", "Add dependencies using the selected dependency scheme")
    ("remove,r", "Remove dependencies using the selected dependency scheme")
    ("outfile,o", boost::program_options::value<std::string>(&out_name), "Write the result to the given file")
    ;

    boost::program_options::options_description hidden_options("Options");
    hidden_options.add_options()
    ("infile,i", boost::program_options::value<std::string>(&in_name), "Input file name")
    ;

    boost::program_options::options_description all_options("All options");
    all_options.add(public_options).add(hidden_options);

    boost::program_options::positional_options_description p;
    p.add("infile", -1);

    boost::program_options::variables_map vm;
    boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(p).run(), vm);
    boost::program_options::notify(vm);

    // Evaluation of parameters
    if (vm.count("help")) {
        std::cout << "Call with\n  " << argv[0] << "<options> <input file>\n\n" << public_options << "\n";
        std::exit(0);
    }

    if (in_name == "") {
        std::cerr << "[ERROR] No input file given." << std::endl;
        std::exit(-1);
    }

    if (vm.count("std") && vm.count("sstd")) {
        std::cerr << "[ERROR] Standard and strict standard dependencies may not be use at the same time." << std::endl;
        std::exit(-1);
    }

    if (vm.count("add") && vm.count("remove")) {
        std::cerr << "[ERROR] Adding and removing dependencies cannot be done at the same time." << std::endl;
        std::exit(-1);
    }

    DependencyOperation operation = DependencyOperation::DO_NOTHING;
    if (vm.count("add")) {
        operation = DependencyOperation::ADD;
    } else if (vm.count("remove")) {
        operation = DependencyOperation::REMOVE;
    }

    Formula dqbf;

    const std::string ext = in_name.substr(in_name.length() - 3, 3);
    if (ext == ".gz" || ext == ".GZ") {
        std::ifstream in(in_name, std::ios_base::in | std::ios_base::binary);
        if (!in) {
            std::cerr << "[ERROR] Could not open input file '" << in_name << "'!" << std::endl;
            std::exit(-1);
        }
        boost::iostreams::filtering_istream in_filter;
        in_filter.push(boost::iostreams::gzip_decompressor());
        in_filter.push(in);
        in_filter >> dqbf;
        in.close();
    } else {
        std::ifstream in(in_name);
        if (!in) {
            std::cerr << "[ERROR] Could not open input file '" << in_name << "'!" << std::endl;
            std::exit(-1);
        }
        in >> dqbf;
        in.close();
    }

    const unsigned int orig_deps = dqbf.numDependencies();

    std::cout << "[INFO] Benchmark: " << in_name << std::endl;
    std::cout << "[INFO] Number of dependencies (before): " << orig_deps << std::endl;

    std::cout << "Manipulating dependencies ..." << std::flush;

    // If removing dependencies and gate detection is enabled, first detect gates.
    if (vm.count("gate") && operation == DependencyOperation::REMOVE) {
        dqbf.applyDependencyScheme(DependencyScheme::GATE, operation);
    }

    // Now apply the requested dependency scheme.
    if (vm.count("std") && !vm.count("rpath")) {
        dqbf.applyDependencyScheme(DependencyScheme::STANDARD, operation);
    } else if (vm.count("sstd") && !vm.count("rpath")) {
        dqbf.applyDependencyScheme(DependencyScheme::STRICT_STANDARD, operation);
    } else if (vm.count("tri") && !vm.count("rpath")) {
        dqbf.applyDependencyScheme(DependencyScheme::REF_TRIANGLE, operation);
    } else if (vm.count("quad") && !vm.count("rpath")) {
        dqbf.applyDependencyScheme(DependencyScheme::REF_QUADRANGLE, operation);
    } else if (vm.count("std")  && vm.count("rpath"))  {
        dqbf.applyDependencyScheme(DependencyScheme::RP_STANDARD, operation);
    } else if (vm.count("sstd") && vm.count("rpath")) {
        dqbf.applyDependencyScheme(DependencyScheme::RP_STRICT_STANDARD, operation);
    } else if (vm.count("tri") && vm.count("rpath")) {
        dqbf.applyDependencyScheme(DependencyScheme::RP_REF_TRIANGLE, operation);
    } else if (vm.count("quad") && vm.count("rpath")) {
        dqbf.applyDependencyScheme(DependencyScheme::RP_REF_QUADRANGLE, operation);
    }

    // If adding dependencies is requested and gate detection enabled, add
    // dependencies using gate detection as the last step.
    if (vm.count("gate") && operation == DependencyOperation::ADD) {
        dqbf.applyDependencyScheme(DependencyScheme::GATE, DependencyOperation::ADD);
    }

    std::cout << "finished." << std::endl;
    std::cout << "[INFO] Number of dependencies (after): " << dqbf.numDependencies() << std::endl;

    if (out_name != "") {
        const std::string ext = out_name.substr(out_name.length() - 3, 3);
        if (ext == ".gz" || ext == ".GZ") {
            std::ofstream out(out_name, std::ios_base::out | std::ios_base::binary);
            if (!out) {
                std::cerr << "[ERROR] Could not open output file '" << out_name << "'!" << std::endl;
                std::exit(-1);
            }
            boost::iostreams::filtering_ostream out_filter;
            out_filter.push(boost::iostreams::gzip_compressor());
            out_filter.push(out);
            out_filter << dqbf;
        } else {
            std::ofstream out(out_name);
            if (!out) {
                std::cerr << "[ERROR] Could not open output file '" << out_name << "'!" << std::endl;
                std::exit(-1);
            }
            out << dqbf;
        }
    }

    return 0;
}
